from django.contrib.auth import user_logged_out
from django.contrib.auth.models import User
from rest_framework import viewsets, permissions, serializers, status
from rest_framework.decorators import action
from rest_framework.response import Response

from .serializers import (
    UserSerializer,
    UserWithDetailSerializer,
    UserLoginSerializer
)


class UserViewSet(viewsets.ModelViewSet):
    queryset = User.objects.select_related(
        'extended_user'
    ).prefetch_related(
        'groups'
    ).all()
    serializer_class = UserSerializer

    @action(detail=True, methods=['post'], permission_classes=[permissions.AllowAny])
    def validate_account(self, request, pk=None) -> Response:
        from .services import validate_account_user
        activation_key = request.POST.get('activation_key')
        message, error = validate_account_user(
            user_pk=pk,
            activation_key=activation_key
        )
        if error:
            raise serializers.ValidationError({'_error': message, 'is_error': error})
        return Response({'result': message})

    @action(detail=True, methods=['post'])
    def add_permission(self, request, pk=None):
        from .services import user_add_permission
        permission_id = int(request.POST.get('permission_id'))
        user = user_add_permission(
            permission_id=permission_id,
            user_id=pk
        )
        serializer = self.get_serializer(user)
        return Response(serializer.data)

    @action(detail=True, methods=['post'])
    def add_group(self, request, pk=None):
        id_group = int(request.POST.get('id_group'))
        from .services import user_add_group
        user = user_add_group(
            grupo_id=id_group,
            user_id=pk
        )
        serializer = self.get_serializer(user)
        return Response(serializer.data)

    @action(detail=True, methods=['post'])
    def change_password(self, request, pk=None):
        from .services import user_change_password
        password_old = request.POST.get('password_old')
        password = request.POST.get('password')
        password_2 = request.POST.get('password_2')
        user_change_password(
            user_id=pk,
            password_new=password,
            password_new_confirmation=password_2,
            password_old=password_old
        )
        return Response({'result': 'La contraseña se ha cambiado correctamente'})

    @action(detail=True, methods=['post'])
    def change_password_other(self, request, pk=None):
        from .services import user_change_password_other
        password = request.POST.get('password')
        password_2 = request.POST.get('password_2')
        user_change_password_other(
            current_user_id=self.request.user.id,
            user_id=pk,
            password_new=password,
            password_new_confirmation=password_2
        )
        return Response({'result': 'La contraseña se ha cambiado correctamente'})

    @action(detail=False, methods=['get'], permission_classes=[permissions.AllowAny])
    def validate_new_username(self, request) -> Response:
        validacion_reponse = {}
        from .services import user_exist_username
        username = self.request.GET.get('username', None)
        if user_exist_username(username=username):
            raise serializers.ValidationError({'username': 'Ya exite'})
        return Response(validacion_reponse)

    @action(detail=False, methods=['post'], permission_classes=[permissions.AllowAny])
    def ask_restore_password(self, request) -> Response:
        from .services import ask_restore_password_account_user
        username = self.request.POST.get('username', None)
        ask_restore_password_account_user(username=username)
        return Response({'result': 'El correo de recuperación de contraseña se ha enviado correctamente'})

    @action(detail=True, methods=['post'], permission_classes=[permissions.AllowAny])
    def set_restored_password(self, request, pk=None) -> Response:
        from .services import set_restored_password
        activation_key = request.POST.get('activation_key')
        password = self.request.POST.get('password', None)
        password_2 = self.request.POST.get('password_2', None)
        message, error = set_restored_password(
            user_pk=pk,
            new_password_2=password_2,
            new_password=password,
            activation_key=activation_key
        )
        if error:
            raise serializers.ValidationError({'_error': message, 'is_error': error})
        return Response({'result': message})

    @action(detail=False, methods=['get'], permission_classes=[permissions.AllowAny])
    def validate_username_login(self, request) -> Response:
        from .services import user_exist_username
        validacion_reponse = {}
        username = self.request.GET.get('username', None)
        if not user_exist_username(username=username):
            raise serializers.ValidationError({'username': 'Este usuario no existe'})
        return Response(validacion_reponse)

    @action(detail=True, methods=['post'])
    def generate_security_qr(self, request, pk=None):
        from .services import user_create_security_qr
        qr = user_create_security_qr(user_id=pk)
        return Response(qr)

    @action(detail=True, methods=['post'])
    def get_security_qr(self, request, pk=None):
        from .services import user_show_security_qr
        qr = user_show_security_qr(user_id=pk)
        return Response(qr)

    @action(detail=True, methods=['post'])
    def change_security_pin(self, request, pk=None):
        from .services import user_change_pin
        pin = request.POST.get('pin')
        password = request.POST.get('password')
        user_change_pin(user_id=pk, password=password, pin=pin)
        return Response({'result': 'El pin se ha cambiado correctamente'})

    @action(detail=True, methods=['post'])
    def change_security_pin_other(self, request, pk=None):
        from .services import user_change_pin_other
        pin = request.POST.get('pin')
        user_change_pin_other(current_user_id=self.request.user.id, user_id=pk, pin=pin)
        return Response({'result': 'El pin se ha cambiado correctamente'})

    @action(detail=True, methods=['post'])
    def upload_profile_image(self, request, pk=None):
        profile_image = self.request.FILES['profile_image']
        user = User.objects.get(pk=pk)
        extended_user = user.extended_user
        extended_user.profile_image = profile_image
        extended_user.save()
        user = User.objects.get(pk=pk)
        serializer = self.get_serializer(user)
        return Response(serializer.data)


class LoginViewSet(viewsets.ModelViewSet):
    serializer_class = UserLoginSerializer
    queryset = User.objects.all()

    @action(detail=False, methods=['post'], permission_classes=[permissions.AllowAny])
    def login(self, request) -> Response:
        from django.http import QueryDict
        username = self.request.POST.get('username')
        password = self.request.POST.get('password')
        import re
        regex = '^\w+([\.-]?\w+)*@\w+([\.-]?\w+)*(\.\w{2,3})+$'
        if re.search(regex, username):
            username = User.objects.values_list('username', flat=True).filter(email__iexact=username)[0]

        ordinary_dict = {'username': username, 'password': password}
        query_dict = QueryDict('', mutable=True)
        query_dict.update(ordinary_dict)
        from .services import user_get_token
        serializer = self.get_serializer(data=query_dict)
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data
        token = user_get_token(user_id=user.id)
        return Response({
            "user": UserWithDetailSerializer(user, context=self.get_serializer_context()).data,
            "token": token
        })

    @action(detail=False, methods=['get'], permission_classes=[permissions.AllowAny])
    def load_user(self, request) -> Response:
        if self.request.user.is_anonymous:
            serializer = UserWithDetailSerializer(None, context={'request': request})
            return Response(serializer.data)
        serializer = UserWithDetailSerializer(self.request.user, context={'request': request})
        return Response(serializer.data)

    @action(detail=False, methods=['get'], permission_classes=[permissions.IsAuthenticated])
    def logout(self, request) -> Response:
        request._auth.delete()
        user_logged_out.send(sender=request.user.__class__,
                             request=request, user=request.user)
        return Response(None, status=status.HTTP_204_NO_CONTENT)

    @action(detail=False, methods=['get'], permission_classes=[permissions.IsAuthenticated])
    def logoutall(self, request) -> Response:
        request.user.auth_token_set.all().delete()
        user_logged_out.send(sender=request.user.__class__,
                             request=request, user=request.user)
        return Response(None, status=status.HTTP_204_NO_CONTENT)



from rest_framework import serializers
from .models import City, Country, State


class CountrySerializer(serializers.ModelSerializer):
    to_string = serializers.SerializerMethodField()
    pk = serializers.IntegerField(required=False)

    def create(self, validated_data):
        validated_data['id'] = validated_data.get('pk')
        return super().create(validated_data)

    def get_to_string(self, obj):
        return obj.name

    class Meta:
        model = Country
        fields = [
            'id',
            'pk',
            'to_string',
            'name',
            'iso3',
            'iso2',
            'phone_code',
            'currency',
            'states'
        ]
        extra_kwargs = {
            'id': {'required': False},
        }
        read_only_fields = ['states']


class StateSerializer(serializers.ModelSerializer):
    country_name = serializers.CharField(source='country.name', read_only=True)
    to_string = serializers.SerializerMethodField()
    pk = serializers.IntegerField(required=False)

    def create(self, validated_data):
        validated_data['id'] = validated_data.get('pk')
        return super().create(validated_data)

    def get_to_string(self, obj):
        return obj.name

    class Meta:
        model = State
        fields = [
            'id',
            'pk',
            'to_string',
            'name',
            'country',
            'country_name',
            'state_code',
            'cities',
        ]
        extra_kwargs = {
            'id': {'required': False},
        }
        read_only_fields = ['cities']


class CitySerializer(serializers.ModelSerializer):
    to_string = serializers.SerializerMethodField()
    pk = serializers.IntegerField(required=False)

    def create(self, validated_data):
        validated_data['id'] = validated_data.get('pk')
        return super().create(validated_data)

    def get_to_string(self, obj):
        return obj.name

    class Meta:
        model = City
        fields = [
            'pk',
            'id',
            'to_string',
            'name',
            'state'
        ]
        extra_kwargs = {
            'id': {'required': False},
        }


class CityWithDetailSerializer(CitySerializer):
    def get_to_string(self, obj):
        return '%s-%s-%s' % (obj.state.country.name, obj.state.name, obj.name)


class CountryWithDetailSerializer(CountrySerializer):
    states = StateSerializer(read_only=True, many=True)


class StateWithDetailSerializer(StateSerializer):
    cities = CitySerializer(many=True, read_only=True)



from django.contrib.auth.models import User, Permission, Group
from django.db.models import Q

from rest_framework import viewsets, permissions
from rest_framework.decorators import action
from rest_framework.response import Response

from utils.permissions.permissions_utilities import DjangoModelPermissionsFull
from .serializers import PermissionSerializer, GroupSerializer


class PermissionViewSet(viewsets.ModelViewSet):
    # permission_classes = [DjangoModelPermissionsFull]
    queryset = Permission.objects.select_related('plus', 'content_type').all()
    serializer_class = PermissionSerializer

    @action(detail=False, methods=['get'])
    def my_permissions(self, request):
        if request.user.is_superuser:
            permissions_list = self.get_queryset()
            serializer = self.get_serializer(permissions_list, many=True)
            return Response(serializer.data)
        permissions_list = self.queryset.filter(
            Q(user=request.user) |
            Q(group__user=request.user)
        ).distinct()
        serializer = self.get_serializer(permissions_list, many=True)
        return Response(serializer.data)

    @action(detail=False, methods=['get'])
    def by_group(self, request):
        group_id = int(self.request.GET.get('group_id'))
        group = Group.objects.get(id=group_id)
        permissions_list = group.permissions.all()
        serializer = self.get_serializer(permissions_list, many=True)
        return Response(serializer.data)

    @action(detail=False, methods=['get'])
    def permiso_x_usuario(self, request):
        user_id = int(request.GET.get('user_id'))
        user = User.objects.get(id=user_id)
        permissions_list = self.queryset.filter(
            Q(user=user) |
            Q(group__user=user)
        ).distinct()
        serializer = self.get_serializer(permissions_list, many=True)
        return Response(serializer.data)

    @action(detail=False, methods=['get'])
    def active(self, request):
        permissions_list = self.queryset.filter(
            plus__active=True
        ).distinct()
        serializer = self.get_serializer(permissions_list, many=True)
        return Response(serializer.data)

    @action(detail=False, methods=['get'], permission_classes=[permissions.IsAuthenticated])
    def has_permissions(self, request):
        permissions_list_parameter = request.GET.get('permissions_list_parameter').split(',')[0:-1]
        if request.user.is_superuser:
            permissions_list = self.queryset.all().filter(codename__in=permissions_list_parameter)
            serializer = self.get_serializer(permissions_list, many=True)
            return Response(serializer.data)
        permissions_list = self.queryset.filter(
            Q(codename__in=permissions_list_parameter) &
            (
                    Q(user=request.user) |
                    Q(group__user=request.user)
            )
        ).distinct()
        serializer = self.get_serializer(permissions_list, many=True)
        return Response(serializer.data)


class GroupViewSet(viewsets.ModelViewSet):
    # permission_classes = [DjangoModelPermissionsFull]
    queryset = Group.objects.all()
    serializer_class = GroupSerializer

    @action(detail=True, methods=['post'])
    def adicionar_permiso(self, request, pk=None):
        grupo = self.get_object()
        id_permiso = int(request.POST.get('id_permiso'))
        permiso = Permission.objects.get(id=id_permiso)
        tiene_permiso = grupo.permissions.filter(id=id_permiso).exists()
        if not tiene_permiso:
            grupo.permissions.add(permiso)
        else:
            grupo.permissions.remove(permiso)

        serializer = self.get_serializer(grupo)
        return Response(serializer.data)

    @action(detail=False, methods=['get'])
    def validar_nombre(self, request) -> Response:
        qs = self.get_queryset()
        validacion_reponse = {}
        name = self.request.GET.get('name', None)
        if name and qs.filter(name=name).exists():
            validacion_reponse.update({'name': 'Ya exite'})
        return Response(validacion_reponse)



# Generated by Django 3.0.5 on 2020-04-22 17:06

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('users_extended', '0007_extendeduser_alias'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='useraditionalpermission',
            options={'managed': False, 'permissions': (('list_user', 'Can list user'), ('change_password_other_user', 'Can change other user password'), ('change_pin_other_user', 'Can change other user pin'), ('change_qr_other_user', 'Can change other user qr'), ('change_profile_other_user', 'Can change other user profile'), ('change_profile_image_other_user', 'Can change other user profile image'), ('change_profile_user', 'Can change user profile'), ('change_password_user', 'Can change user password'), ('change_pin_user', 'Can change user pin'), ('change_qr_user', 'Can change user qr'), ('change_profile_image_user', 'Can change user profile image'), ('change_permission_user', 'Can change user permission'), ('make_superuser_user', 'Can make user superuser'), ('make_staff_user', 'Can make user staff'), ('make_active_user', 'Can make user active'))},
        ),
    ]



# Generated by Django 3.0.5 on 2020-04-02 23:47

import app_configurations.models
from django.db import migrations, models
import imagekit.models.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='GeneralInfoConfiguration',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('application_name', models.CharField(default='App. Name', max_length=50)),
                ('logo', imagekit.models.fields.ProcessedImageField(blank=True, null=True, upload_to=app_configurations.models.GeneralInfoConfiguration.imagen_logo_upload_to)),
            ],
        ),
    ]



from .models import State, City, Country


def create_multiple_countries(
        countries
):
    countries_to_create = []
    for kwards in countries:
        country = Country()
        for field in kwards:
            value = kwards[field]
            if hasattr(country, field):
                setattr(country, field, value)
        countries_to_create.append(country)

    if len(countries_to_create) > 0:
        Country.objects.bulk_create(countries_to_create, batch_size=1000)


def create_multiple_states(
        states
):
    states_to_create = []
    for kwards in states:
        state = State()
        for field in kwards:
            value = kwards[field]
            if hasattr(state, field):
                setattr(state, field, value)
        states_to_create.append(state)

    if len(states_to_create) > 0:
        State.objects.bulk_create(states_to_create)


def create_multiple_cities(
        cities
):
    cities_to_create = []
    for kwards in cities:
        city = City()
        for field in kwards:
            value = kwards[field]
            if hasattr(city, field):
                setattr(city, field, value)
        cities_to_create.append(city)

    if len(cities_to_create) > 0:
        City.objects.bulk_create(cities_to_create, batch_size=1000)



from django.http import Http404
from rest_framework.permissions import DjangoModelPermissions
from rest_framework import exceptions

SAFE_METHODS = ('GET', 'HEAD', 'OPTIONS', 'VIEW')


class DjangoModelPermissionsFull(DjangoModelPermissions):
    perms_map = {
        'GET': ['%(app_label)s.list_%(model_name)s'],
        'VIEW': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': [],
        'HEAD': [],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }

    def has_permission(self, request, view):
        # Workaround to ensure DjangoModelPermissions are not applied
        # to the root view when using DefaultRouter.
        if getattr(view, '_ignore_model_permissions', False):
            return True

        if not request.user or (
                not request.user.is_authenticated and self.authenticated_users_only):
            return False

        queryset = self._queryset(view)
        perms = self.get_required_permissions(request.method, queryset.model)
        return request.user.has_perms(perms)

    def get_required_object_permissions(self, method, model_cls):
        kwargs = {
            'app_label': model_cls._meta.app_label,
            'model_name': model_cls._meta.model_name
        }

        if method not in self.perms_map:
            raise exceptions.MethodNotAllowed(method)

        return [perm % kwargs for perm in self.perms_map[method]]

    def has_object_permission(self, request, view, obj):
        if request.method != 'GET':
            return True
        queryset = self._queryset(view)
        model_cls = queryset.model
        user = request.user
        perms = self.get_required_object_permissions('VIEW', model_cls)
        if not user.has_perms(perms):
            if request.method in SAFE_METHODS:
                raise Http404
            read_perms = self.get_required_object_permissions('VIEW', model_cls)
            if not user.has_perms(read_perms, obj):
                raise Http404
            return False
        return True



from django.db.models.signals import pre_delete, post_init, post_save
from django.dispatch import receiver

from .models import ExtendedUser


@receiver(pre_delete, sender=ExtendedUser)
def profile_image_pre_delete(sender, instance, **kwargs):
    instance.profile_image.delete(False)


@receiver(post_init, sender=ExtendedUser)
def backup_profile_image_path(sender, instance, **kwargs):
    instance._current_profile_image = instance.profile_image


@receiver(post_save, sender=ExtendedUser)
def delete_profile_image(sender, instance, **kwargs):
    if hasattr(instance, '_current_profile_image'):
        if instance._current_profile_image != instance.profile_image:
            instance._current_profile_image.delete(save=False)



from django.contrib.auth.models import Permission, Group
from rest_framework import serializers

from utils.serializers.serializers_utilities import CustomSerializerMixin
from .models import PermissionPlus


class PermissionSerializer(CustomSerializerMixin, serializers.ModelSerializer):
    name_plus = serializers.CharField(source='plus.name', allow_null=True, allow_blank=True)
    active = serializers.BooleanField(source='plus.active')
    to_string = serializers.SerializerMethodField()
    content_type_label = serializers.CharField(source='content_type.app_label', read_only=True)
    content_type_model = serializers.CharField(source='content_type.app_label', read_only=True)

    def get_to_string(self, instance):  # pragma: no cover
        return instance.plus.name.title() if hasattr(
            instance,
            'plus'
        ) and instance.plus.name else instance.name.title()

    class Meta:
        model = Permission
        fields = [
            'id',
            'to_string',
            'codename',
            'content_type',
            'content_type_label',
            'content_type_model',
            'name',
            'name_plus',
            'active',
        ]

    def update(self, instance, validated_data):
        plus_data = validated_data.pop('plus', None)
        if hasattr(instance, 'plus'):
            plus = instance.plus
        else:
            plus = PermissionPlus.objects.create(**plus_data)

        for attr, value in plus_data.items():
            setattr(plus, attr, value)
        plus.permission = instance
        plus.save()
        return super().update(instance, validated_data)


class GroupSerializer(serializers.ModelSerializer):
    to_string = serializers.SerializerMethodField()

    def get_to_string(self, instance):  # pragma: no cover
        return instance.name.title()

    class Meta:
        model = Group
        fields = [
            'id',
            'name',
            'permissions',
            'to_string'
        ]



from rest_framework import routers
from .views import (
    CountryViewSet,
    StateViewSet,
    CityViewSet
)

router = routers.DefaultRouter()
router.register('countries', CountryViewSet)
router.register('countries_states', StateViewSet)
router.register('countries_cities', CityViewSet)



import operator
import random
import string
from functools import reduce
from django.conf import settings
from cryptography.fernet import Fernet, MultiFernet
from django.db import models


def fernet_base() -> MultiFernet:
    key = settings.KEY_ENCRY
    vector = settings.VECTOR_ENCRY
    key1 = Fernet(key)
    key2 = Fernet(vector)
    fernet = MultiFernet([key1, key2])
    return fernet


def generate_secret_word(size, pre=None, pos=None) -> str:
    characters = list(string.ascii_letters + "-._~")
    word = ''
    for x in range(size):
        word += random.choice(characters)
    if pre:
        word = str(pre) + word
    if pos:
        word += str(pos)
    return word


def encrypt(texto_plano: str) -> str:
    if not texto_plano:  # pragma: no cover
        return ''
    fernet = fernet_base()
    new_text = fernet.encrypt(str.encode(texto_plano))
    return new_text.decode()


def decrypt(enc: str) -> str:
    if not enc:  # pragma: no cover
        return ''
    fernet = fernet_base()
    decrypted_text = fernet.decrypt(str.encode(enc))
    return decrypted_text.decode()


def search_multiple_fields(queryset, search_fields, parameter):
    def construct_search(field_name):
        if field_name.startswith('^'):
            return "%s__istartswith" % field_name[1:]
        elif field_name.startswith('='):
            return "%s__iexact" % field_name[1:]
        elif field_name.startswith('@'):
            return "%s__search" % field_name[1:]
        else:
            return "%s__icontains" % field_name

    orm_lookups = [construct_search(str(search_field))
                   for search_field in search_fields]
    for bit in parameter.split():
        or_queries = [models.Q(**{orm_lookup: bit})
                      for orm_lookup in orm_lookups]
        queryset = queryset.filter(reduce(operator.or_, or_queries))
    return queryset



from rest_framework import routers
from .views import (
    PermissionViewSet,
    GroupViewSet
)

router = routers.DefaultRouter()
router.register(r'permissions', PermissionViewSet)
router.register(r'permissions_groups', GroupViewSet)



from datetime import timedelta

from django.contrib.auth.models import User, Permission, Group
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.utils import timezone
from knox.models import AuthToken
from rest_framework import serializers
from django.conf import settings
from utils.functions.functions import encrypt, decrypt, generate_secret_word
import crypt
from users_extended.models import ExtendedUser


def send_registration_email(user: User, email: str):
    subject, from_email, to = 'Validación Registro', '%s <%s>' % (
        settings.SITE_NAME, settings.DEFAULT_FROM_EMAIL), email
    activation_key = decrypt(user.extended_user.registration_token)
    context = {
        "domain": settings.MAIN_DOMAIN,
        "user": user,
        "activation_key": activation_key
    }
    text_content = render_to_string('emails/registration_email.html', context=context)
    msg = EmailMultiAlternatives(subject, text_content, from_email, [to])
    msg.attach_alternative(text_content, "text/html")
    msg.send()


def ask_restore_password_account_user(
        username: str
):
    user = None
    import re
    regex = '^\w+([\.-]?\w+)*@\w+([\.-]?\w+)*(\.\w{2,3})+$'
    if re.search(regex, username):
        if User.objects.filter(email__iexact=username).exists():
            user = User.objects.filter(email__iexact=username)[0]
    else:
        if User.objects.filter(username__iexact=username).exists():
            user = User.objects.filter(username__iexact=username)[0]
    if user is not None:
        if hasattr(user, 'extended_user'):
            extended_user = user.extended_user
        else:
            extended_user = ExtendedUser.objects.create(user=user)
        now = timezone.now()
        word = generate_secret_word(30)
        extended_user.registration_token = encrypt(word)
        extended_user.registration_expiration = now + timedelta(days=2)
        extended_user.save()

        subject, from_email, to = 'Validación Registro', '%s <%s>' % (
            settings.SITE_NAME,
            settings.DEFAULT_FROM_EMAIL
        ), user.email
        activation_key = decrypt(extended_user.registration_token)
        context = {
            "domain": settings.MAIN_DOMAIN,
            "user": user,
            "activation_key": activation_key
        }
        text_content = render_to_string('emails/restore_password.html', context=context)
        msg = EmailMultiAlternatives(subject, text_content, from_email, [to])
        msg.attach_alternative(text_content, "text/html")
        msg.send()

    else:
        raise serializers.ValidationError({'_error': 'El usuario no existe'})


def create_update_user(
        user: User = None,
        **validated_data
) -> User:
    registration = validated_data.pop('registration') if 'registration' in validated_data else False
    extended_user_values = validated_data.pop('extended_user') if 'extended_user' in validated_data else None
    validated_data.pop('groups', None)
    validated_data.pop('user_permissions', None)
    email = validated_data.get('email', None)
    now = timezone.now()
    exist_email_qs = User.objects.filter(email__iexact=email)
    if user is not None:
        exist_email_qs = exist_email_qs.exclude(pk=user.pk)
    if exist_email_qs.exists():
        raise serializers.ValidationError({'email': 'Ya existe un usuario con este email, utilizar otro'})
    if user is None:
        user = User.objects.create_user(is_active=not registration, **validated_data)
    else:
        for attr, value in validated_data.items():
            if hasattr(user, attr):
                setattr(user, attr, value)
        if registration:
            user.is_active = False
        user.save()
    if not hasattr(user, 'extended_user'):
        extended_user = ExtendedUser.objects.create(user_id=user.id)
    else:
        extended_user = user.extended_user
    if extended_user_values is not None:
        for attr, value in extended_user_values.items():
            if hasattr(extended_user, attr):
                setattr(extended_user, attr, value)
    if registration:
        extended_user.registration_token = encrypt(generate_secret_word(40))
        extended_user.registration_expiration = now + timedelta(days=7)
    extended_user.save()
    user.refresh_from_db()
    if registration:
        send_registration_email(user, email)
    return user


def validate_account_user(
        activation_key: str,
        user_pk: int
) -> [str, bool]:
    user = User.objects.get(pk=user_pk)
    was_activated = user.is_active
    same_activation_key = False
    expired_link = True
    if not was_activated:
        if hasattr(user, 'extended_user'):
            same_activation_key = decrypt(user.extended_user.registration_token) == activation_key
        if same_activation_key:
            expired_link = timezone.now().date() > user.extended_user.registration_expiration
            if expired_link:
                message = 'El link de activación para "%s" ha caducado, registrase de nuevo.' % user.username
            else:
                extended_user = user.extended_user
                extended_user.registration_token = None
                extended_user.registration_expiration = None
                extended_user.save()
                user.is_active = True
                user.save()
                message = 'Se ha activado correctamente la cuenta de "%s".' % user.username
        else:
            message = 'El token de activación para "%s" no coincide.' % user.username
    else:
        same_activation_key = True
        expired_link = False
        message = 'La cuenta para "%s" ya se había activado.' % user.username
    error = not same_activation_key or expired_link
    return message, error


def set_restored_password(
        activation_key: str,
        user_pk: int,
        new_password: str,
        new_password_2: str,
) -> [str, bool]:
    if new_password != new_password_2:
        message = 'La contraseña y su confirmación no coinciden, por favor verificar las dos.'
        return message, True

    user = User.objects.get(pk=user_pk)
    extended_user = user.extended_user
    if extended_user.registration_token is None or extended_user.registration_expiration is None:
        message = 'Este link ya ha sido utilizado, o no sirve para restaurar contraseña'
        return message, True

    same_activation_key = decrypt(user.extended_user.registration_token) == activation_key
    expired_link = timezone.now().date() > user.extended_user.registration_expiration
    if same_activation_key:
        if expired_link:
            message = 'El link de cambio de contraseña para "%s" ha caducado, solicitelo de nuevo.' % user.username
        else:
            extended_user.registration_token = None
            extended_user.registration_expiration = None
            extended_user.save()
            user.set_password(new_password)
            user.save()
            message = 'Se ha cambiado correctamente la contraseña de "%s".' % user.username
    else:
        message = 'El token de cambio de contraseña para "%s" no coincide.' % user.username
    error = not same_activation_key or expired_link
    return message, error


def set_security_pin(user_id: int, raw_pin: str) -> [User, str]:
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        extended_user = user.extended_user
        extended_user.security_pin = crypt.crypt(raw_pin)
        extended_user.save()
        return user, extended_user.security_pin
    else:
        return None, None


def user_validate_security_pin(user_id: int, raw_pin: str):
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        return crypt.crypt(raw_pin, user.extended_user.security_pin) == user.extended_user.security_pin
    return False


def user_get_token(
        user_id: int
) -> str:
    user = User.objects.get(pk=user_id)
    now = timezone.now()
    expired_tokens = user.auth_token_set.filter(expiry__lt=now)
    expired_tokens.delete()
    tokens = user.auth_token_set.all()
    if hasattr(user, 'extended_user'):
        extended_user = user.extended_user
    else:
        extended_user = ExtendedUser.objects.create(user=user)
    if not extended_user.multi_sessions:
        tokens.delete()
    else:
        qty_sessions_allowed = extended_user.multi_sessions_qty
        if qty_sessions_allowed > 0:
            already_open = tokens.count()
            if qty_sessions_allowed - already_open == 0:
                tokens.order_by('created')[0].delete()
    _, token = AuthToken.objects.create(user)
    return token


def user_exist_username(
        username: str
) -> bool:
    import re
    regex = '^\w+([\.-]?\w+)*@\w+([\.-]?\w+)*(\.\w{2,3})+$'
    if re.search(regex, username):
        if User.objects.filter(email__iexact=username).exists():
            return True
    else:
        if User.objects.filter(username=username).exists():
            return True
    return False


def user_validate_security_qr(user_id: int, raw_qr: str):
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        return decrypt(user.extended_user.security_qr) == raw_qr
    return False


def user_show_security_qr(user_id: int) -> str:
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        return decrypt(user.extended_user.security_qr)
    return ''


def user_create_security_qr(user_id: int) -> str:
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        new_qr = generate_secret_word(30, user.id, user.id)
        extended_user = user.extended_user
        encrypted_new_qr = encrypt(new_qr)
        extended_user.security_qr = encrypted_new_qr
        extended_user.save()
        return new_qr
    return ''


def user_change_pin(
        user_id: int,
        password: str,
        pin: str
) -> User:
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        if not user.check_password(password):
            raise serializers.ValidationError(
                {'_error': 'La contraseña suministrada no coincide con el usuario, es necesario para cambiar el PIN'})
        user = set_security_pin(
            user_id=user_id,
            raw_pin=pin
        )
    return user


def user_change_pin_other(
        current_user_id: int,
        user_id: int,
        pin: str
) -> User:
    current_user = User.objects.get(pk=current_user_id)
    change_pin_other_user = current_user.has_perm('users_extended.change_pin_other_user')
    if not change_pin_other_user:
        raise serializers.ValidationError({'_error': 'No tiene permiso de cambiar el pin de otro usuario'})
    user = User.objects.get(pk=user_id)
    if hasattr(user, 'extended_user'):
        user = set_security_pin(
            user_id=user_id,
            raw_pin=pin
        )
    return user


def user_change_password(
        user_id: int,
        password_old: str,
        password_new: str,
        password_new_confirmation: str
) -> User:
    user = User.objects.get(pk=user_id)
    if not user.check_password(password_old):
        raise serializers.ValidationError({'_error': 'La contraseña suministrada no coincide con el user'})
    if not password_new == password_new_confirmation:
        raise serializers.ValidationError({'_error': 'La contraseña nueva con su confirmación no coincide'})
    user.set_password(password_new)
    user.save()
    return user


def user_change_password_other(
        current_user_id: int,
        user_id: int,
        password_new: str,
        password_new_confirmation: str
) -> User:
    current_user = User.objects.get(pk=current_user_id)
    user = User.objects.get(pk=user_id)
    change_password_other_user = current_user.has_perm('users_extended.change_password_other_user')
    if not change_password_other_user:
        raise serializers.ValidationError({'_error': 'No tiene permiso de cambiar la contraseña de otro usuario'})
    if not password_new == password_new_confirmation:
        raise serializers.ValidationError({'_error': 'La contraseña nueva no coincide con su confirmación'})
    user.set_password(password_new)
    user.save()
    return user


def user_add_permission(
        permission_id: int,
        user_id: int
) -> User:
    user = User.objects.get(pk=user_id)
    permission = Permission.objects.get(id=permission_id)
    has_group = user.user_permissions.filter(id=permission_id).exists()
    if not has_group:
        user.user_permissions.add(permission)
    else:
        user.user_permissions.remove(permission)
    return user


def user_add_group(
        grupo_id: int,
        user_id: int
) -> User:
    user = User.objects.get(pk=user_id)
    group = Group.objects.get(id=grupo_id)

    has_group = user.groups.filter(id=grupo_id).exists()
    if not has_group:
        user.groups.add(group)
    else:
        user.groups.remove(group)
    return user



# Generated by Django 3.0.5 on 2020-05-04 18:54

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('countries', '0006_auto_20200504_1407'),
        ('users_extended', '0010_auto_20200426_0245'),
    ]

    operations = [
        migrations.AddField(
            model_name='extendeduser',
            name='city',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='countries.City'),
        ),
    ]
